﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Problems.BinPacking.Interfaces;
using HeuristicLab.Operators;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Common;
using HeuristicLab.Data;
using HeuristicLab.Problems.BinPacking.Decoders;
using HeuristicLab.Parameters;
using HeuristicLab.Problems.BinPacking.Shapes;
using HeuristicLab.Encodings.PackingEncoding.PackingPlan;

namespace HeuristicLab.Problems.BinPacking.Evaluators {
  [Item("Packingplan Evaluation Algorithm", "Represents a composition of a decoder and an evaluator for bin-packing problems.")]
  [StorableClass]
  public class PackingPlanEvaluationAlgorithm<E, D, B, I> : SingleSuccessorOperator, IPackingPlanEvaluationAlgorithm
    where E : Item //E defines the used Encoding
    where D : class, IPackingDimensions
    where B : PackingShape<D>, IPackingBin
    where I : PackingShape<D>, IPackingItem {

    public IValueLookupParameter<IPackingSolutionDecoder> PackingSolutionDecoderParameter {
      get { return (IValueLookupParameter<IPackingSolutionDecoder>)Parameters["PackingSolutionDecoder"]; }
    }
    ILookupParameter<IPackingSolutionDecoder> IPackingPlanEvaluationAlgorithm.PackingSolutionDecoderParameter {
      get { return PackingSolutionDecoderParameter; }
    }
    public IValueLookupParameter<IPackingPlanEvaluator> PackingPlanEvaluatorParameter {
      get { return (IValueLookupParameter<IPackingPlanEvaluator>)Parameters["PackingPlanEvaluator"]; }
    }
    ILookupParameter<IPackingPlanEvaluator> IPackingPlanEvaluationAlgorithm.PackingPlanEvaluatorParameter {
      get { return PackingPlanEvaluatorParameter; }
    }
    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }

    [StorableConstructor]
    protected PackingPlanEvaluationAlgorithm(bool deserializing) : base(deserializing) { }
    protected PackingPlanEvaluationAlgorithm(PackingPlanEvaluationAlgorithm<E, D, B, I> original, Cloner cloner)
      : base(original, cloner) {
    }
    public PackingPlanEvaluationAlgorithm ()
      : base() {
        Parameters.Add(new ValueLookupParameter<IPackingSolutionDecoder>("PackingSolutionDecoder", "The decoding operator that is used to calculate a packing plan from the used representation."));
        Parameters.Add(new ValueLookupParameter<IPackingPlanEvaluator>("PackingPlanEvaluator", "The actual packing plan evaluation operator."));
        Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The quality value aka fitness value of the solution."));
        QualityParameter.Hidden = true;
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new PackingPlanEvaluationAlgorithm<E, D, B, I>(this, cloner);
    }

    public override IOperation Apply() {
      var decoder = PackingSolutionDecoderParameter.ActualValue;
      var evaluator = PackingPlanEvaluatorParameter.ActualValue;
      if (evaluator == null) throw new InvalidOperationException("A PackingPlanEvaluator could not be found.");

      var operations = new OperationCollection(base.Apply());
      operations.Insert(0, ExecutionContext.CreateChildOperation(evaluator));
      if (decoder != null) // decode before evaluating
        operations.Insert(0, ExecutionContext.CreateChildOperation(decoder));
      return operations;
    }
  }
}